import numpy as np
from ..actors.esp import Esp
from ..world.world import World
from ..world.mode import Mode
from ..world.point import Point
from ..world.spline import Spline
from ..helpers import eliminate_pi
from .navigation.carrot import Carrot


def ramp(x: float, in_min: float, in_max: float, out_min: float, out_max: float, clip: bool = False):
    if clip and x < in_min:
        return out_min
    if clip and x > in_max:
        return out_max
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min


async def drive_spline(spline: Spline, world: World, esp: Esp, *, flip_hook: bool = False, throttle_at_end: bool = True):
    if spline.start.distance(spline.end) < 0.01:
        return  # NOTE: skip tiny splines

    hook_offset = Point(x=world.robot.parameters.hook_offset, y=0) * (-1 if flip_hook else 1)
    carrot_offset = Point(x=world.robot.parameters.carrot_offset, y=0)
    carrot = Carrot(spline=spline, offset=carrot_offset)

    while True:
        hook = world.robot.prediction.transform(hook_offset)
        if not carrot.move(hook, distance=world.robot.parameters.carrot_distance):
            break
        world.robot.carrot = carrot.pose

        turn_angle = eliminate_pi(hook.direction(carrot.offset_point) - world.robot.prediction.yaw)
        curvature = np.tan(turn_angle) / hook_offset.x
        if curvature != 0 and abs(1 / curvature) < world.robot.parameters.minimum_turning_radius:
            curvature = (-1 if curvature < 0 else 1) / world.robot.parameters.minimum_turning_radius

        drive_backward = hook.projected_distance(carrot.offset_point, world.robot.prediction.yaw) < 0
        linear = -1 if drive_backward else 1
        if carrot.t > 1.0 and throttle_at_end:
            linear *= ramp(carrot.target_distance, 0.5, 0.0, 1.0, 0.5)
        angular = linear * curvature

        await esp.drive(*throttle(world, linear, angular))

    world.robot.carrot = None
    await esp.drive(0, 0)


def throttle(world: World, linear: float, angular: float):
    if world.mode != Mode.TEST:  # TODO: require camera tracking in tests as well
        if world.robot.parameters.max_detection_age_ramp is None:
            factor = 1
        elif world.robot.detection is None:
            factor = 0
        else:
            age_ramp = world.robot.parameters.max_detection_age_ramp
            age = world.time - world.robot.detection.time
            factor = ramp(age, age_ramp[0], age_ramp[1], 1.0, 0.0, clip=True)
        linear *= factor
        angular *= factor

    linear_limit = world.robot.parameters.linear_speed_limit
    angular_limit = world.robot.parameters.angular_speed_limit

    if abs(linear) > linear_limit:
        factor = linear_limit / abs(linear)
        angular *= factor
        linear *= factor
    if abs(angular) > angular_limit:
        factor = angular_limit / abs(angular)
        linear *= factor
        angular *= factor

    return linear, angular



import numpy as np
from ..actors.esp import Esp
from ..world.world import World


async def drive_arc(world: World, esp: Esp):
    while world.robot.prediction.x < 2:
        await esp.drive(1, np.deg2rad(25))

    await esp.drive(0, np.deg2rad(0))



from __future__ import annotations
from pydantic import BaseModel
import abc
from .velocity import Velocity

from typing import TYPE_CHECKING
if TYPE_CHECKING:  # NOTE: avoid cyclic import
    from .world import World


class HardwareGroup(BaseModel, abc.ABC):
    name: str
    output: bool = False

    @abc.abstractproperty
    def commands(self) -> list[str]:
        pass

    def with_output(self):
        self.output = True
        return self

    def parse(self, words: list[str], world: World):
        pass


class Bluetooth(HardwareGroup):
    device_name: str

    def __init__(self, **data):
        super().__init__(name='bluetooth', **data)

    @property
    def commands(self) -> list[str]:
        return [
            f'new bluetooth {self.name} ESP_{self.device_name}',
        ]


class Can(HardwareGroup):
    rxPin: str
    txPin: str

    def __init__(self, **data):
        super().__init__(name='can', **data)

    @property
    def commands(self) -> list[str]:
        return [
            f'new can {self.name} {self.rxPin},{self.txPin}',
        ]


class Led(HardwareGroup):
    pin: str
    interval: float = 0.1
    duty: float = 0.5
    repeat: bool = True

    @property
    def commands(self) -> list[str]:
        return [
            f'new led {self.name} {self.pin}',
            f'set {self.name}.interval={self.interval}',
            f'set {self.name}.duty={self.duty}',
            f'set {self.name}.repeat={"1" if self.repeat else "0"}',
        ]


class Button(HardwareGroup):
    pin: str

    @property
    def commands(self) -> list[str]:
        return [
            f'new button {self.name} {self.pin}',
        ]


class RoboClawWheels(HardwareGroup):
    address: int = 128
    baud: int = 38400
    m_per_tick: float
    width: float

    @property
    def commands(self) -> list[str]:
        return [
            f'new roboclawwheels {self.name} {self.address},{self.baud}',
            f'set {self.name}.mPerTick={self.m_per_tick}',
            f'set {self.name}.width={self.width}',
        ]

    def parse(self, words: list[str], world: World):
        world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=world.robot.hardware_time,
        ))
        world.robot.temperature = float(words.pop(0))
        world.robot.battery = float(words.pop(0))


class ODriveMotor(HardwareGroup):
    can: Can
    device_id: int
    m_per_tick: float

    @property
    def commands(self) -> list[str]:
        return [
            f'new odrivemotor {self.name} 0,{self.can.name},{hex(self.device_id)[2:]}',
            f'set {self.name}.mPerTick={self.m_per_tick}',
        ]


class ODriveWheels(HardwareGroup):
    left: ODriveMotor
    right: ODriveMotor
    width: float
    left_power_factor: float = 1.0
    right_power_factor: float = 1.0

    @property
    def commands(self) -> list[str]:
        return self.left.commands + self.right.commands + [
            f'new odrivewheels {self.name} {self.left.name},{self.right.name}',
            f'set {self.name}.width={self.width}',
            f'set {self.name}.leftPowerFactor={self.left_power_factor}',
            f'set {self.name}.rightPowerFactor={self.right_power_factor}',
        ]

    def parse(self, words: list[str], world: World):
        world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=world.robot.hardware_time,
        ))



import numpy as np
from ..world.point import Point
from ..world.spline import Spline
from ..world.world import World
from ..actors.esp import Esp
from ..helpers import eliminate_2pi
from .spline import drive_spline, throttle


async def drive_path(world: World, esp: Esp):
    for s, path_segment in enumerate(world.path):
        spline = path_segment.spline
        if s == 0 or spline.start.distance(world.path[s-1].spline.end) > 0.01:
            await drive_to(world, esp, spline.start)
        is_last_segment = path_segment == world.path[-1]
        await drive_spline(spline, world, esp, throttle_at_end=is_last_segment, flip_hook=path_segment.backward)


async def drive_to(world: World, esp: Esp, target: Point):
    if world.robot.parameters.minimum_turning_radius:
        await drive_circle(world, esp, target)

    robot_pose = world.robot.prediction
    approach_spline = Spline(
        start=robot_pose.point,
        control1=robot_pose.point.interpolate(target, 1/3),
        control2=robot_pose.point.interpolate(target, 2/3),
        end=target,
    )
    await drive_spline(approach_spline, world, esp)


async def drive_circle(world: World, esp: Esp, target: Point):
    while True:
        direction = world.robot.prediction.point.direction(target)
        angle = eliminate_2pi(direction - world.robot.prediction.yaw)
        if abs(angle) < np.deg2rad(5):
            break
        linear = 0.5
        sign = 1 if angle > 0 else -1
        angular = linear / world.robot.parameters.minimum_turning_radius * sign
        await esp.drive(*throttle(world, linear, angular))



from typing import Coroutine
from ..world.world import World, WorldState
from .actor import Actor


class Automator(Actor):
    interval: float = 0.1

    def __init__(self) -> None:
        super().__init__()
        self.routines = []

    def add(self, coro: Coroutine):
        self.routines.append(coro)

    def replace(self, coro: Coroutine):
        self.routines.clear()
        self.add(coro)

    async def step(self):
        if self.world.state != WorldState.RUNNING:
            return

        for coro in self.routines:
            try:
                coro.send(None)
            except StopIteration:
                self.routines.remove(coro)
                if not self.routines:
                    await self.pause_automations(because='the last one has completed')
            except:
                await self.pause_automations(because='an exception occurred in an automation')
                self.routines.clear()
                self.log.exception(f'paused and cleared automations due to exception in {coro}')



import aioserial
from functools import reduce
from operator import ixor
from ..world.world import World
from .esp import Esp


class SerialEsp(Esp):
    interval: float = 0.01

    def __init__(self):
        super().__init__()
        aioserial.AioSerial('/dev/esp', baudrate=115200)  # NOTE try to open serial port (factory needs this)
        self.aioserial = None
        self.remainder = ''
        self.last_step = None

    async def step(self):
        dt = self.world.time - self.last_step if self.last_step is not None else 0
        if dt > 1:
            msg = 'esp serial communication can not be guaranteed (>= 1 sec)'
            self.log.error(msg + '; aborting automations')
            await self.pause_automations(because=msg)
        elif dt > 0.1:
            self.log.warn('esp serial communication is slow (>= 100 ms)')
        self.last_step = self.world.time

        if not self.is_open():
            return

        try:
            # NOTE we just read the existing bytes from buffer... no need to go async
            self.remainder += self.aioserial.read_all().decode()
        except:
            self.log.warning('Error reading from serial')
            return

        self.remainder = await self.parse(self.remainder, self.world)

    async def send_async(self, line):
        if not self.is_open():
            return
        line = f'{line}^{reduce(ixor, map(ord, line))}\n'
        try:
            # HACK writing synchronous because the async way contains a lock which does not work with our pause/resume of automations
            self.aioserial.write(line.encode())
        finally:
            await self.sleep(0)  # make sure we let other coroutines do their work on the event loop

    def is_open(self):
        if self.aioserial is None:
            self.aioserial = aioserial.AioSerial('/dev/esp', baudrate=115200)
        return self.aioserial.isOpen()

    def connect(self):
        if not self.is_open():
            return
        self.aioserial.open()

    def disconnect(self):
        if not self.is_open():
            return
        self.aioserial.close()

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_open else "disconnected"})'



import time
from operator import ixor
from functools import reduce
from .. import task_logger
from ..world.world import World
from ..world.hardware import HardwareGroup
from .actor import Actor
from .. import event


class Esp(Actor):

    async def drive(self, linear: float, angular: float):
        await self.send_async('wheels speed %.3f,%.3f' % (linear, angular))

    async def power(self, left: float, right: float):
        await self.send_async('wheels power %.3f,%.3f' % (left, right))

    def configure(self, hardware: list[HardwareGroup]):
        task_logger.create_task(self.configure_async(hardware))

    async def configure_async(self, hardware: list[HardwareGroup]):
        await self.send_async('esp erase')

        for group in hardware:
            for command in group.commands:
                await self.send_async('+' + command)
                time.sleep(0.1)

        output_modules = ','.join(group.name for group in hardware if group.output)
        await self.send_async(f'+set esp.outputModules={output_modules}')
        await self.send_async('+esp unmute')
        await self.send_async('+set esp.ready=1')
        await self.send_async('+set esp.24v=1')
        await self.send_async('esp restart')

    async def parse(self, messages: str, world: World) -> str:
        '''Parses the messages received from the ESP microcontroller and writes the data into the world.
        Note: An incomplete message at the end is not parsed but returned to be completed later.'''

        millis = None
        while '\n' in messages:
            line, messages = messages.split('\n', 1)
            if line.startswith("\x1b[0;32m"):
                self.log.warning(line)
                continue  # NOTE: ignore green log messages
            if '^' in line:
                line, checksum = line.split('^')
                if reduce(ixor, map(ord, line)) != int(checksum):
                    self.log.warning('Checksum failed')
                    continue
            if not line.startswith("esp "):
                self.log.warning(line)
                continue  # NOTE: ignore all messages but esp status
            try:
                words = line.split()[1:]
                millis = float(words.pop(0))
                if world.robot.clock_offset is None:
                    continue
                world.robot.hardware_time = millis / 1000 + world.robot.clock_offset
                for group in world.robot.hardware:
                    if group.output:
                        group.parse(words, world)
            except (IndexError, ValueError):
                self.log.warning(f'Error parsing esp message "{line}"')

        if millis is not None:
            world.robot.clock_offset = world.time - millis / 1000

        await event.call(event.Id.NEW_MACHINE_DATA)
        return messages



import socketio
import socket
from ..world.world import World
from .esp import Esp


class WebEsp(Esp):
    '''Remote connection to the Robot Brains Esp.

    This makes it possible to keep developing on your fast computer 
    while communicating with the hardware components connected to a physical Robot Brain.
    '''
    interval: float = 0.01

    def __init__(self):
        super().__init__()
        self.host = 'ws://192.168.43.1:80'
        # let's try to establish a connection to the esp proxy synchronously...
        self._try_socketio()
        # ... ok, we passed without exception -> creating async client
        self.sio = socketio.AsyncClient()

        @self.sio.event
        def connect():
            assert self.sio.transport() == 'websocket'
            self.log.info('connected to esp proxy')

        @self.sio.event
        async def read(data):
            await self.parse(data + '\n', self.world)

    async def step(self):
        if not self.sio.connected:
            await self.sio.connect(self.host)

    async def send_async(self, line):
        await self.sio.emit('write', line)

    async def tear_down(self):
        await self.sio.disconnect()

    def _try_socketio(self):

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        host = self.host.replace('ws://', '',).split(':')
        s.connect(host)
        s.shutdown(2)

        sio = socketio.Client(reconnection=False,)

        @sio.event
        def connect():
            sio.disconnect()

        sio.connect(self.host)



from pydantic import BaseModel
from typing import Optional
from ...world.point import Point
from ...world.pose import Pose
from ...world.spline import Spline


class Carrot(BaseModel):
    spline: Spline
    offset: Point = Point(x=0, y=0)
    t: float = 0
    target_distance: Optional[float] = None

    @property
    def pose(self) -> Pose:
        if self.t < 1.0:
            return self.spline.pose(self.t)
        else:
            return Pose(
                x=self.spline.x(1.0) + (self.t - 1.0) * self.spline.gx(1.0),
                y=self.spline.y(1.0) + (self.t - 1.0) * self.spline.gy(1.0),
                yaw=self.spline.yaw(1.0),
            )

    @property
    def offset_point(self) -> Point:
        return self.pose.transform(self.offset)

    def move(self, hook: Point, distance: float = 1.0, move_threshold: float = 0.01):
        end_pose = self.spline.pose(1.0)
        end_point = end_pose.point
        end_yaw = end_pose.yaw

        while hook.distance(self.offset_point) < distance:
            self.t += 0.01
            if self.t < 1.0:
                continue
            self.target_distance = hook.projected_distance(end_point, end_yaw)
            if self.target_distance <= move_threshold:
                return False

        return True



from nicegui.ui import Ui
from rosys.actors.esp import Esp
from rosys.actors.serial_esp import SerialEsp
from rosys import event


class LizardSerialDebug():
    ui: Ui = None  # will be set by rosys.ui.configure
    lizard: Esp = None  # will be set by rosys.ui.configure

    def __init__(self) -> None:
        if type(self.lizard) is not SerialEsp:
            return
        self.ui.switch('Lizard', value=self.lizard.is_open(), on_change=self.change)

    async def change(self, status):
        if status.value:
            self.lizard.connect()
            await event.call(event.Id.NEW_NOTIFICATION, 'connected to Lizard')
        else:
            await event.call(event.Id.PAUSE_AUTOMATIONS, 'lizard is deactivated')
            self.lizard.disconnect()
            await event.call(event.Id.NEW_NOTIFICATION, 'disconnected from to Lizard')



#!/usr/bin/env python3
from nicegui import ui
import numpy as np
import pylab as pl
import time
from binary_renderer import BinaryRenderer

renderer = BinaryRenderer((30, 40))

x, y, r = 10, 15, 5
t = time.time()
renderer.circle(x, y, r)
ui.label(f"circle:  {(time.time() - t) * 1000:.3f} ms")

points = np.array([[20.5, 10], [30, 15], [25, 25]])
t = time.time()
renderer.polygon(points)
ui.label(f"polygon:  {(time.time() - t) * 1000:.3f} ms")

with ui.plot():
    pl.imshow(renderer.map, cmap=pl.cm.gray, interpolation='nearest')
    pl.gca().add_patch(pl.matplotlib.patches.Circle((x, y), r, color='none', ec='C0'))
    pl.fill(points[:, 0], points[:, 1], color='none', ec='C0')

ui.run()



import asyncio
import logging
from typing import Callable
from concurrent.futures import ProcessPoolExecutor

from rosys.world.world import World
from .. import event


class Actor:
    world: World
    interval: float = None
    process_pool = ProcessPoolExecutor()

    def __init__(self) -> None:
        self.name = __name__[:-5] + self.__class__.__name__
        self.log = logging.getLogger(self.name)

    async def step(self):
        pass

    async def tear_down(self):
        pass

    async def sleep(self, seconds: float):
        '''delay execution; in tests this method will be replaced'''
        await asyncio.sleep(seconds)

    async def run_io_bound(self, callback: Callable, *args: any):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, callback, *args)

    async def run_cpu_bound(self, callback: Callable, *args: any):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, callback, *args)

    async def pause_automations(self, *, because: str):
        await event.call(event.Id.PAUSE_AUTOMATIONS, because)

    async def notify(self, message: str):
        await event.call(event.Id.NEW_NOTIFICATION, message)

    def __str__(self) -> str:
        return type(self).__name__



from __future__ import annotations  # NOTE: for PEP 563 (postponed evaluation of annotations)
from pydantic import BaseModel
import numpy as np
from .point import Point
from dataclasses import dataclass


@dataclass
class PoseStep():
    linear: float
    angular: float
    time: float


class Pose(BaseModel):
    x: float = 0
    y: float = 0
    yaw: float = 0
    time: float = 0

    @property
    def point(self) -> Point:
        return Point(x=self.x, y=self.y)

    @property
    def matrix(self):
        return np.array([
            [np.cos(self.yaw), -np.sin(self.yaw), self.x],
            [np.sin(self.yaw), np.cos(self.yaw), self.y],
            [0, 0, 1],
        ])

    @property
    def yaw_deg(self):
        return np.rad2deg(self.yaw)

    @staticmethod
    def from_matrix(M, time: float = 0):
        return Pose(x=M[0, -1], y=M[1, -1], yaw=np.arctan2(M[1, 0], M[0, 0]), time=time)

    def __str__(self):
        return '%.3f, %.3f, %.1f deg' % (self.x, self.y, self.yaw_deg)

    def distance(self, other: Pose) -> float:
        return self.point.distance(other.point)

    def projected_distance(self, other: Pose) -> float:
        return self.point.projected_distance(other.point, other.yaw)

    def __iadd__(self, step: PoseStep):
        self.x += step.linear * np.cos(self.yaw)
        self.y += step.linear * np.sin(self.yaw)
        self.yaw += step.angular
        self.time = step.time
        return self

    def transform(self, point: Point) -> Point:
        return Point(
            x=self.x + point.x * np.cos(self.yaw) - point.y * np.sin(self.yaw),
            y=self.y + point.x * np.sin(self.yaw) + point.y * np.cos(self.yaw),
        )

    def transform_pose(self, pose: Pose) -> Pose:
        return Pose(
            x=self.x + pose.x * np.cos(self.yaw) - pose.y * np.sin(self.yaw),
            y=self.y + pose.x * np.sin(self.yaw) + pose.y * np.cos(self.yaw),
            yaw=self.yaw + pose.yaw,
        )

    def relative_pose(self, other: Pose) -> Pose:
        return Pose.from_matrix(np.linalg.inv(self.matrix) @ other.matrix)

    def rotate(self, angle: float) -> Pose:
        return Pose(x=self.x, y=self.y, yaw=self.yaw+angle, time=self.time)



from __future__ import annotations
from pydantic import BaseModel
import numpy as np


class Rotation(BaseModel):
    R: list[list[float]]

    @staticmethod
    def zero():
        return Rotation(R=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    @staticmethod
    def from_euler(omega: float, phi: float, kappa: float) -> Rotation:
        Rx = np.array([[1, 0, 0], [0, np.cos(omega), -np.sin(omega)], [0, np.sin(omega), np.cos(omega)]])
        Ry = np.array([[np.cos(phi), 0, np.sin(phi)], [0, 1, 0], [-np.sin(phi), 0, np.cos(phi)]])
        Rz = np.array([[np.cos(kappa), -np.sin(kappa), 0], [np.sin(kappa), np.cos(kappa), 0], [0, 0, 1]])
        return Rotation(R=(Rz @ Ry @ Rx).tolist())

    def __mul__(self, other) -> Rotation:
        return Rotation(R=np.dot(self.R, other.R).tolist())

    @property
    def T(self) -> Rotation:
        return Rotation(R=np.array(self.R).T.tolist())

    @property
    def euler(self) -> tuple[float]:
        return (
            np.arctan2(self.R[2][1], self.R[2][2]),
            np.arctan2(-self.R[2][0], np.sqrt(self.R[2][1]**2 + self.R[2][2]**2)),
            np.arctan2(self.R[1][0], self.R[0][0]),
        )

    @property
    def total_angle(self) -> float:
        # https://en.wikipedia.org/wiki/Rotation_matrix#Determining_the_angle
        return np.arccos((np.trace(self.R) - 1) / 2)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return '%6.1f %6.1f %6.1f' % tuple(np.round(np.rad2deg(self.euler), 2))



import asyncio
from asyncio.tasks import Task
import collections
import inspect
from aenum import Enum, auto
from typing import Awaitable, Callable, Union
import logging
import weakref

tasks: list[Task] = []
listeners = collections.defaultdict(set)
log = logging.getLogger('rosys.event')


class Id(Enum, init='value __doc__'):
    '''Event Identifier. Every event has its own set of parameters.'''

    def _generate_next_value_(name, start, count, last_values):
        '''uses enum name as value when calling auto()'''
        return name

    NEW_MACHINE_DATA = auto(), 'triggered in high frequency whenever machine data had been read'
    PAUSE_AUTOMATIONS = auto(), 'call this event to pause any running automations; provide a description of the cause as string parameter.'
    NEW_NOTIFICATION = auto(), 'call this event to notify the user; provide the message as string parameter'


def register(event: Id, listener: Union[Callable, Awaitable]):
    if not callable(listener):
        raise Exception('non-callable listener')
    if listener.__name__ == "<lambda>":  # NOTE lambda functions must be stored without weakref because they will be collected otherwise
        ref = listener
    elif inspect.ismethod(listener):
        ref = weakref.WeakMethod(listener)
    else:
        ref = weakref.ref(listener)
    listeners[event].add(ref)


def unregister(event: Id, listener: Union[Callable, Awaitable]):
    listeners[event].remove(listener)


async def call(event: Id, *args):
    '''Fires event and waits async until all registered listeners are completed'''

    for listener in list(listeners.get(event, {})):
        try:
            if hasattr(listener, '__name__') and listener.__name__ == '<lambda>':
                listener(*args)
                continue
            if listener() is None:
                unregister(event, listener)
                continue
            if inspect.iscoroutinefunction(listener()):
                await listener()(*args)
            else:
                listener()(*args)
        except:
            log.exception(f'could not call {listener=} for {event=}')


def emit(event: Id, *args):
    '''Fires event without waiting for the result.'''

    loop = asyncio.get_event_loop()
    for listener in list(listeners.get(event, {})):
        try:
            if hasattr(listener, '__name__') and listener.__name__ == '<lambda>':
                tasks.append(loop.run_in_executor(None, listener, *args))
                continue
            if listener() is None:
                unregister(event, listener)
                continue
            if inspect.iscoroutinefunction(listener()):
                tasks.append(loop.create_task(listener()(*args), name=f'handle {event=}'))
            else:
                tasks.append(loop.run_in_executor(None, listener,  *args))
        except:
            log.exception(f'could not call {listener=} for {event=}')



import numpy as np
from ..helpers import angle
from ..world.pose import PoseStep
from ..world.world import World
from .actor import Actor
from .. import event


class Odometer(Actor):

    def __init__(self):
        super().__init__()

        self.last_time: float = None
        self.steps: list[PoseStep] = []
        self.flips: int = 0
        self.flip_detection_initialized: bool = False
        event.register(event.Id.NEW_MACHINE_DATA, self.handle_velocity)

    def handle_velocity(self):
        if not self.world.robot.odometry:
            return
        while self.world.robot.odometry:
            velocity = self.world.robot.odometry.pop(0)

            if self.last_time is None:
                self.last_time = velocity.time
                continue

            dt = velocity.time - self.last_time
            self.last_time = velocity.time

            step = PoseStep(linear=dt*velocity.linear, angular=dt*velocity.angular, time=self.world.time)
            self.steps.append(step)
            self.world.robot.prediction += step
            self.world.robot.simulation += step

            self.world.robot.current_velocity = velocity

            if step.linear or step.angular:
                self.world.robot.last_movement = step.time

        self.prune_steps(self.world.time - 10.0)

    def handle_detection(self):
        if self.world.robot.detection is None or not any(self.steps) or self.world.robot.detection.time < self.steps[0].time:
            return

        while self.steps[0].time < self.world.robot.detection.time:
            del self.steps[0]

        # NOTE: attempt to avoid 180-degree flips due to swapped marker points
        if self.flip_detection_initialized:
            dYaw = sum(step.angular for step in self.steps)
            if abs(angle(self.world.robot.prediction.yaw - dYaw, self.world.robot.detection.yaw)) > np.deg2rad(90):
                self.flips += 1
                for image in self.world.images:
                    if image.time == self.world.robot.detection.time:
                        self.log.warning(f'adding {image.id} to upload queue because our position has flipped')
                        self.world.upload.mark(image)
                if self.flips < 3:
                    self.log.warn('Avoiding flip')
                    return
            else:
                self.flips = 0

        self.world.robot.prediction = self.world.robot.detection.copy(deep=True)
        for step in self.steps:
            self.world.robot.prediction += step

        self.flip_detection_initialized = True

    def prune_steps(self, cut_off_time: float):
        while any(self.steps) and self.steps[0].time < cut_off_time:
            del self.steps[0]



import asyncio
from typing import Callable
from rosys.runtime import Runtime
from rosys.world.mode import Mode
from rosys.world.robot import Robot
from rosys.world.world import World, WorldState


class TestRuntime(Runtime):
    __test__ = False
    is_time_running: bool = True

    def __init__(self, world: World = None):
        if world is None:
            world = World(mode=Mode.TEST, state=WorldState.RUNNING, robot=Robot())
        super().__init__(world)
        self.world.set_time(0)  # NOTE in tests we start at zero for better reading

        from .helper import set_global_runtime  # NOTE import here to avoid PytestAssertRewriteWarning
        set_global_runtime(self)
        self.exception = None

    async def forward(self, seconds, dt=0.01):
        # NOTE we start runtime here because this makes it easy in the tests to prepare it beforehand
        if not self.tasks:
            for actor in self.actors:
                actor.sleep = self.sleep
                actor.run_cpu_bound = self.run_cpu_bound
            await self.start()

        end_time = self.world.time + seconds
        self.log.info(f'-------------------> forwarding to {round(end_time,2)}')
        while self.world.time <= end_time:
            if self.is_time_running:
                self.world.set_time(self.world.time + dt)
                await asyncio.sleep(0)
            else:
                await asyncio.sleep(0.01)
            if self.exception is not None:
                raise RuntimeError(f'error while forwarding time {dt} s') from self.exception

        self.log.info(f'-------------------> now it\'s {round(self.world.time,2)}')

    async def sleep(self, seconds: float):
        sleep_end_time = self.world.time + seconds
        while self.world.time <= sleep_end_time:
            await asyncio.sleep(0)

    async def run_cpu_bound(self, callback: Callable, *args: any):
        try:
            self.is_time_running = False
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, callback, *args)
        finally:
            self.is_time_running = True

    def handle_exception(self, ex: Exception):
        super().handle_exception(ex)
        self.exception = ex



import logging
import json
import os
from .world.obstacle import Obstacle
from .world.path_segment import PathSegment
from .world.robot import RobotParameters
from .world.world import World


class Persistence:

    def __init__(self, world: World, filepath: str = '~/.rosys/world.json'):
        self.world = world
        self.filepath = os.path.expanduser(filepath)

    def dump(self) -> dict:
        return {
            'path': [path_segment.dict() for path_segment in self.world.path],
            'robot': {'parameters': self.world.robot.parameters.dict()},
            'obstacles': {id: obstacle.dict() for id, obstacle in self.world.obstacles.items()},
        }

    def load(self, dict: dict):
        self.world.path = [PathSegment.parse_obj(path_segment) for path_segment in dict['path']]
        self.world.robot.parameters = RobotParameters.parse_obj(dict['robot']['parameters'])
        self.world.obstacles = {id: Obstacle.parse_obj(obstacle) for id, obstacle in dict['obstacles'].items()}

    def backup(self):
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)
        with open(self.filepath, 'w') as f:
            json.dump(self.dump(), f)

    def restore(self):
        if not os.path.exists(self.filepath):
            logging.warning('No backup file found.')
            return
        try:
            with open(self.filepath, 'r') as f:
                self.load(json.load(f))
        except:
            logging.exception('Could not load from backup')



import inspect
import time
import traceback
import sys
import numpy as np


def measure(*, reset: bool = False, ms: bool = False):
    global t
    if 't' in globals() and not reset:
        dt = time.time() - t
        line = inspect.stack()[1][0].f_lineno
        output = f'{dt * 1000:7.3f} ms' if ms else f'{dt:7.3f} s'
        print(f'{inspect.stack()[1].filename}:{line}', output, flush=True)
    if reset:
        print('------------', flush=True)
    t = time.time()


def print_stacktrace():
    print("-" * 60)
    traceback.print_exc(file=sys.stdout)
    print("-" * 60, flush=True)


def angle(yaw0: float, yaw1: float) -> float:
    return eliminate_2pi(yaw1 - yaw0)


def eliminate_pi(angle: float) -> float:
    return (angle + np.pi / 2) % np.pi - np.pi / 2


def eliminate_2pi(angle: float) -> float:
    return (angle + np.pi) % (2 * np.pi) - np.pi



from nicegui.ui import Ui
import logging
import yappi
from tabulate import tabulate
import asyncio

log = logging.getLogger('rosys.profiler')


def create_profiler(ui: Ui):

    async def start():
        ui.notify('start profiling')
        profile_button.props(replace='icon=stop')
        yappi.clear_stats()
        yappi.start()
        for i in range(100):
            await asyncio.sleep(0.1)
            if not yappi.is_running():
                return  # profiling was stopped early
        stop()

    def stop():
        ui.notify('stop profiling')
        profile_button.props(replace='icon=play_arrow')
        yappi.stop()
        table = [
            [str(v) for v in [stat.full_name, stat.ttot, stat.tsub, stat.tavg, stat.ncall]]
            for stat in yappi.get_func_stats()
            if 'python' not in stat.module
        ]
        output = tabulate(
            table[:15],
            headers=['function', 'total', 'excl. sub', 'avg', 'ncall'],
            floatfmt=".4f"
        )
        print(output, flush=True)
        yappi.get_thread_stats().print_all()

    async def toggle() -> bool:
        if yappi.is_running():
            stop()
        else:
            await start()
        return False  # do not refresh UI

    profile_button = ui.button('Profiler', on_click=toggle).props('icon=play_arrow')



from nicegui.ui import Ui

from .. import Runtime
from .. import task_logger
from ..automations import drive_path
from ..world.world import WorldState


class AutomationControls:
    ui: Ui
    runtime: Runtime

    def __init__(self) -> None:
        def pause():
            task_logger.create_task(self.runtime.pause())
            self.runtime.world.tracking = False

        def toggle_automation():
            if self.runtime.world.state == WorldState.PAUSED:
                if not self.runtime.automator.routines:
                    self.runtime.automator.add(drive_path(self.runtime.world, self.runtime.esp))
                self.runtime.resume()
                self.runtime.world.tracking = True
            elif self.runtime.world.state == WorldState.RUNNING:
                pause()

        def stop():
            pause()
            self.runtime.automator.routines.clear()

        automation_button = self.ui.button(on_click=toggle_automation).props('icon=play_arrow outline')
        self.ui.button(on_click=stop).props('icon=stop outline')

        def refresh_steering():
            if self.runtime.world.state == WorldState.RUNNING:
                automation_button.props(replace='icon=pause')
            if self.runtime.world.state == WorldState.PAUSED:
                automation_button.props(replace='icon=play_arrow')
        self.ui.timer(0.1, refresh_steering)



import asyncio
import numpy as np
from ..world.velocity import Velocity
from ..world.world import World
from .esp import Esp
from .. import event


class MockedEsp(Esp):
    interval: float = 0.01

    def __init__(self):
        super().__init__()

        x = [point[0] for point in self.world.robot.shape.outline]
        self.width = max(x) - min(x)
        self.linear_velocity: float = 0
        self.angular_velocity: float = 0

    async def step(self):
        velocity = Velocity(linear=self.linear_velocity, angular=self.angular_velocity, time=self.world.time)
        self.world.robot.odometry.append(velocity)
        self.world.robot.battery = 25.0 + np.sin(0.1 * self.world.time) + 0.02 * np.random.randn()
        self.world.robot.temperature = np.random.uniform(34, 35)
        await event.call(event.Id.NEW_MACHINE_DATA)

    async def send_async(self, line):
        if line.startswith("wheels power "):
            left = float(line.split()[2].split(',')[0])
            right = float(line.split()[2].split(',')[1])
            self.linear_velocity = (left + right) / 2.0
            self.angular_velocity = (right - left) / self.width / 2.0

        if line.startswith("wheels speed "):
            self.linear_velocity = float(line.split()[2].split(',')[0])
            self.angular_velocity = float(line.split()[2].split(',')[1])

        await asyncio.sleep(0)



from pydantic import BaseModel, PrivateAttr
from enum import Enum
import time
from .mode import Mode
from .obstacle import Obstacle
from .path_segment import PathSegment
from .robot import Robot


class WorldState(Enum):
    RUNNING = 1
    PAUSED = 2


class World(BaseModel):
    robot: Robot = Robot()
    mode: Mode = Mode.REAL
    state: WorldState = WorldState.PAUSED
    _time: float = PrivateAttr(default_factory=time.time)
    tracking: bool = False
    robot_locator_cam: str = None
    download_queue: list[str] = []
    image_data: dict[str, bytes] = {}
    link_queue: list[list[str]] = []
    path: list[PathSegment] = []
    obstacles: dict[str, Obstacle] = {}
    notifications: list[tuple[float, str]] = []

    @property
    def time(self):
        return self._time if self.mode == Mode.TEST else time.time()

    def set_time(self, value):
        assert self.mode == Mode.TEST
        self._time = value



from typing import Union
import pytest
import numpy as np
from rosys.runtime import Runtime
from rosys.automations.drive_path import drive_to
from rosys.world.point import Point
from rosys.world.point3d import Point3d

global_runtime: Runtime = None


def set_global_runtime(runtime: Runtime):
    global global_runtime
    global camera_count
    global_runtime = runtime
    camera_count = 0


def assert_pose(x: float, y: float, *, deg: float = None, linear_tolerance: float = 0.1, deg_tolerance: float = 1.0):
    pose = global_runtime.world.robot.prediction
    assert pose.x == pytest.approx(x, abs=linear_tolerance)
    assert pose.y == pytest.approx(y, abs=linear_tolerance)

    if deg is not None:
        assert np.rad2deg(pose.yaw) == pytest.approx(deg, abs=deg_tolerance)


def assert_point(actual: Union[Point, Point3d], expected: Union[Point, Point3d], tolerance=0.1):
    assert type(actual) == type(expected)

    assert actual.x == pytest.approx(expected.x, abs=tolerance)
    assert actual.y == pytest.approx(expected.y, abs=tolerance)

    if type(actual) is Point3d:
        assert actual.z == pytest.approx(expected.z, abs=tolerance)


def automate_drive_to(x: float, y: float):
    global_runtime.automator.replace(drive_to(global_runtime.world, global_runtime.esp, Point(x=x, y=y)))



from nicegui.ui import Ui
from nicegui import app
from ..runtime import Runtime
from .joystick import Joystick as joystick
from .keyboard_control import KeyboardControl as keyboard_control
from .obstacle_object import ObstacleObject as obstacle_object
from .path_object import PathObject as path_object
from .robot_object import RobotObject as robot_object
from .automation_controls import AutomationControls as automation_controls
from .lizard_serial_debug import LizardSerialDebug as lizard_serial_debug
from .objgraph_page import ObjgraphPage as objgraph_page
from .pyloot_page import PylootPage as pyloot_page
from .. import event


def configure(ui: Ui, runtime: Runtime):
    ui.on_startup(runtime.start())
    ui.on_shutdown(runtime.stop())
    event.register(event.Id.NEW_NOTIFICATION, ui.notify)

    joystick.steerer = runtime.steerer
    keyboard_control.ui = ui
    keyboard_control.steerer = runtime.steerer
    path_object.world = runtime.world
    robot_object.robot = runtime.world.robot
    robot_object.ui = ui
    automation_controls.runtime = runtime
    automation_controls.ui = ui
    lizard_serial_debug.ui = ui
    lizard_serial_debug.lizard = runtime.esp
    objgraph_page.ui = ui
    pyloot_page.ui = ui
    pyloot_page.app = app



from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Curve
from rosys.world.world import World


class PathObject(Object3D):
    world: World = None  # will be set by rosys.ui.configure

    def __init__(self):
        super().__init__('group')
        self.update()

    def update(self) -> bool:
        [obj.delete() for obj in list(self.view.objects.values()) if obj.name == 'path']
        for segment in self.world.path:
            Curve(
                [segment.spline.start.x, segment.spline.start.y, 0],
                [segment.spline.control1.x, segment.spline.control1.y, 0],
                [segment.spline.control2.x, segment.spline.control2.y, 0],
                [segment.spline.end.x, segment.spline.end.y, 0],
            ).material('#ff8800').with_name('path')
        return False  # NOTE: avoid JustPy page_update



import numpy as np
from nicegui.ui import Ui
from nicegui.events import KeyEventArguments
from rosys.world.point import Point
from rosys.actors.steerer import Steerer


class KeyboardControl:
    steerer: Steerer = None  # will be set by rosys.ui.configure
    ui: Ui = None  # will be set by rosys.ui.configure

    def __init__(self, *, default_speed: float = 2.0):
        self.ui.keyboard(on_key=self.handle_keys)
        self.direction = Point(x=0, y=0)
        self.speed = default_speed

    def handle_keys(self, e: KeyEventArguments):
        if e.action.repeat:
            return

        # change speed via number key
        if e.action.keydown and e.key.number is not None:
            self.speed = e.key.number
            self.direction.y = self.speed * np.sign(self.direction.y)
            self.direction.x = self.speed * np.sign(self.direction.x)
            self.steerer.update(self.direction.y, self.direction.x)

        # update individual speed components via arrow keys
        if e.modifiers.shift and e.key.is_cursorkey:
            if e.action.keydown and self.direction.x == 0 and self.direction.y == 0:
                self.steerer.start()

            new_speed = self.speed if e.action.keydown else 0
            if e.key.arrow_left:
                self.direction.y = -new_speed
            elif e.key.arrow_right:
                self.direction.y = +new_speed
            elif e.key.arrow_up:
                self.direction.x = +new_speed
            elif e.key.arrow_down:
                self.direction.x = -new_speed
            self.steerer.update(self.direction.y, self.direction.x)

            if e.action.keyup and self.direction.x == 0 and self.direction.y == 0:
                self.steerer.stop()

        # stop when releasing shift key
        if e.key.shift and e.action.keyup:
            self.direction.y = 0
            self.direction.x = 0
            self.steerer.stop()



import numpy as np
from rosys.actors.pathplanning.grid import Grid
from rosys.actors.pathplanning.obstacle_map import ObstacleMap
from rosys.actors.pathplanning.robot_renderer import RobotRenderer


def generate_experiment(id_):
    group = abs(int(id_))
    variant = int(np.round(abs(id_) - group, 1) * 10)
    sign = 1 if id_ >= 0 else -1

    grid = Grid((60, 80, 36), (0.0, 0.0, 8.0, 6.0))

    robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)

    if group == 1:
        obstacles = [
            [1.5, 1.5, 0.2, 6.0],
            [1.5, 1.5, 5.0, 0.2],
            [6.3, 1.5, 0.2, 2.0],
            [4.5, 2.5, 0.2, 1.5],
            [5.5, 5.0, 0.2, 0.2],
        ]
        pose = (0.2, 0.75, 0)
        goal = (3.0, 3.5, 0) if variant == 0 else (0.75, 4.0, -np.pi / 2)
        obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)
        small_obstacle_map = None

    if group == 2:
        obstacles = [
            [0.0, 2.0, 3.0, 0.2],
            [4.3, 2.0, 4.0, 0.2],
        ]
        pose = (0.5, 1.0, 0)
        goal = (3.6, 3.5, np.pi / 2)
        obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)
        small_obstacle_map = None

    if group == 4:
        obstacles = [
            [5.0, 2.31, 3.0, 0.2],
            [5.0, 3.70, 3.0, 0.2],
        ]
        pose = (1.0, 3.1, 0)
        goal = (7.0, 3.1, np.pi)
        obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)
        small_obstacle_map = None

    if group == 5:
        obstacles = [
            [4.0, 2.0, 2.0, 1.0],
            [4.0, 4.0, 2.0, 1.0],
        ]
        pose = (1.0, 1.0, 0 if variant == 0 else np.pi)
        goal = (6.0, 3.5, 0 if variant == 0 else np.pi)
        obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)
        small_obstacle_map = None

    backward_to_goal = sign == -1

    return robot_renderer, pose, goal, obstacle_map, small_obstacle_map, backward_to_goal



from .actors.pathplanning.planner import Planner
from .persistence import Persistence
from .runtime import Runtime
from .world.mode import Mode
from .world.obstacle import Obstacle
from .world.path_segment import PathSegment
from .world.point import Point
from .world.point3d import Point3d
from .world.pose import Pose
from .world.robot import Robot, RobotShape
from .world.spline import Spline
from .world.world import World, WorldState
from . import event



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rosys',
 'rosys.actors',
 'rosys.actors.pathplanning',
 'rosys.automations',
 'rosys.automations.navigation',
 'rosys.test',
 'rosys.ui',
 'rosys.world']

package_data = \
{'': ['*']}

install_requires = \
['aenum>=3.1.5,<4.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'aioserial>=1.3.0,<2.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'more-itertools>=8.10.0,<9.0.0',
 'networkx>=2.6.2,<3.0.0',
 'nicegui>=0.6.7',
 'numpy>=1.20.1,<2.0.0',
 'objgraph>=3.5.0,<4.0.0',
 'opencv-contrib-python-headless>=4.5.4,<5.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyloot>=0.0.7,<0.0.8',
 'pyserial>=3.5,<4.0',
 'python-socketio[asyncio_client]>=5.3.0,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'retry>=0.9.2,<0.10.0',
 'scipy>=1.7.2,<2.0.0',
 'sh>=1.14.2,<2.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'yappi>=1.3.3,<2.0.0']

setup_kwargs = {
    'name': 'rosys',
    'version': '0.4.8',
    'description': 'Modular Robot System With Elegant Automation Capabilities',
    'long_description': "# RoSys - The Robot System\n\nRoSys provides an easy-to-use robot system.\nIts purpose is simlar to [ROS](https://www.ros.org/).\nBut RoSys is fully based on modern web technologies and focusses on mobile robotics.\n\nSee full documentation at [rosys.io](https://rosys.io/).\n\n## Features\n\n**All Python**\n: Business logic is wired in Python while computation-heavy tasks are encapsulated through websockets or bindings.\n\n**Shared State**\n: All code can access and manipulate a shared and typesafe state -- this does not mean it should.\nGood software design is still neccessary.\nBut it is much easier to do if you do not have to perform serialization all the time.\n\n**No Threading**\n: Thanks to [asyncio](https://docs.python.org/3/library/asyncio.html) you can write the business logic without locks and mutex mechanisms.\nThe running system feels like everything is happening in parallel.\nBut each code block is executed one after another through an event queue and yields execution as soon as it waits for I/O or heavy computation.\nThe latter is still executed in threads to not block the rest of the business logic.\n\n**Web UI**\n: Most machines need some kind of human interaction.\nWe made sure your robot can be operated fully off the grid with any web browser by incorporating [NiceGUI](https://nicegui.io/).\nIt is also possible to proxy the user interface through a gateway for remote operation.\n\n**Simulation**\n: Robot hardware is often slower than your own computer.\nTherefore RoSys supports a simulation mode for rapid development.\nTo get maximum performance the current implementation does not run a full physics engine.\n\n**Testing**\n: You can use pytest to write high-level integration tests.\nIt is based on the above-described simulation mode and accelerates the robot's time for super fast execution.\n",
    'author': 'Zauberzeug GmbH',
    'author_email': 'info@zauberzeug.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zauberzeug/rosys',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



from ..actors.esp import Esp
from ..world.world import World
from ..world.point import Point
from .drive_path import drive_to


async def drive_square(world: World, esp: Esp):
    start_pose = world.robot.prediction.copy()
    for x, y in [(1, 0), (1, 1), (0, 1), (0, 0)]:
        await drive_to(world, esp, start_pose.transform(Point(x=x, y=y)))



#!/usr/bin/env python3
from nicegui import ui
import pylab as pl
import plot_tools as pt
import time
from uuid import uuid4
from rosys.world.obstacle import Obstacle
from rosys.world.point import Point
from rosys.world.pose import Pose
from rosys.world.robot import Robot
from rosys.world.mode import Mode
from rosys.world.world import World
from planner import Planner
import icecream
icecream.install()

world = World(mode=Mode.SIMULATION, robot=Robot())
for x, y, w, h in [
    [1.5, 1.5, 0.2, 6.0],
    [1.5, 1.5, 5.0, 0.2],
    [6.3, 1.5, 0.2, 2.0],
    [4.5, 2.5, 0.2, 1.5],
    [5.5, 5.0, 0.2, 0.2],
]:
    id = str(uuid4())
    outline = [Point(x=x, y=y), Point(x=x+w, y=y), Point(x=x+w, y=y+h), Point(x=x, y=y+h)]
    world.obstacles[id] = Obstacle(id=id, outline=outline)

planner = Planner(world)

plot = ui.plot()


def run():
    t = time.time()
    planner.search(goal=Pose(x=10, y=4, yaw=0))
    print("path finding: %5.1f ms" % ((time.time() - t) * 1000), flush=True)

    with plot:
        pl.clf()
        pt.show_distance_map(planner.distance_map)
        pt.show_obstacle_map(planner.obstacle_map)
        pl.gca().invert_yaxis()
        pl.autoscale(False)
        pt.plot_path(planner.raw_path, 'C0')
        [pl.plot(s.target[0], s.target[1], 'C0.') for s in planner.raw_path]
        pt.plot_path(planner.path, 'C2', lw=2)
        [pl.plot(s.target[0], s.target[1], 'C2o') for s in planner.path]
        for step in planner.path:
            pt.plot_robot(world.robot.shape, (step.spline.x(1), step.spline.y(1), step.spline.yaw(1)), 'C2', lw=1)


run()
ui.button('Run', on_click=run)

ui.run()



#!/usr/bin/env python3
from nicegui import ui
import pylab as pl
import numpy as np
import time
from grid import Grid
from robot_renderer import RobotRenderer
from obstacle_map import ObstacleMap
from distance_map import DistanceMap
import plot_tools as pt

grid = Grid((60, 80, 36), (0, 0, 16.0, 12.0))
obstacles = [
    [0.0, 6.0, 5.6, 0.4],
    [10.8, 6.0, 7.2, 0.4],
]
robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)
obstacle_map = ObstacleMap.from_list(grid, obstacles, robot_renderer)

target = 4.0, 2.0

t = time.time()
distance_map = DistanceMap(obstacle_map, target)
ui.label("%5.3f ms" % ((time.time() - t) * 1000))

with ui.plot():
    pt.show_distance_map(distance_map)
    pt.show_obstacle_map(obstacle_map)
    extent = pt.bbox_to_extent(grid.bbox)

with ui.plot():
    rows = np.arange(0, grid.size[0] - 1, 0.2)
    cols = np.arange(0, grid.size[1] - 1, 0.2)
    xx, yy = grid.from_grid(rows, cols)
    interp = distance_map.interpolate(xx, yy).reshape(len(rows), len(cols))
    pl.imshow(interp, cmap=pl.cm.gray, interpolation='nearest', extent=extent, clim=[0, 30])
    pt.show_obstacle_map(obstacle_map)
    pl.autoscale(False)
    for y in np.linspace(8.0, 11.0, 4):
        for x in np.linspace(2.0, 14.0, 30):
            dx, dy = distance_map.gradient(x, y)
            pl.plot(x, y, 'C2.', ms=3)
            pl.plot([x, x + dx], [y, y + dy], 'C2', lw=1)

with ui.plot():
    Gx = distance_map.gradient(xx, yy)[0].reshape(len(yy), len(xx))
    pl.imshow(Gx, cmap=pl.cm.gray, interpolation='nearest', extent=extent, clim=[-1, 1])
    pt.show_obstacle_map(obstacle_map)

ui.run()



from nicegui.elements.log import Log
from nicegui.ui import Ui
import rosys
import rosys.ui
from rosys import World
from rosys import event
from datetime import datetime
import objgraph
import random
import base64


class ObjgraphPage:
    ui: Ui = None  # will be set by rosys.ui.configure

    def __init__(self) -> None:
        with self.ui.page('/objgraph'):
            self.ui.label('Analyze Memory')
            self.ui.button('log most common types', on_click=lambda: objgraph.show_most_common_types(limit=20))
            self.ui.button('log growth', on_click=lambda: objgraph.show_growth(limit=20))

            self.ui.input('Search Object', on_change=self.update_graph)
            self.graph = self.ui.row().style('width:40%')
            self.graph.visible = False

    def update_graph(self, event):
        objgraph.show_chain(objgraph.find_backref_chain(
            random.choice(objgraph.by_type(event.value)),  # just pick one of the found objects
            objgraph.is_proper_module),
            filename='/tmp/rosys_objgraph.png')
        with open('/tmp/rosys_objgraph.png', "rb") as f:
            image = base64.b64encode(f.read()).decode('utf-8')
        self.graph.clear()
        self.graph.visible = True
        with self.graph:
            self.ui.image('data:image/png;base64,'+str(image))



from .actors.esp import Esp
from .actors.serial_esp import SerialEsp
from .actors.mocked_esp import MockedEsp
from .actors.web_esp import WebEsp
from .world.world import World
import logging

log = logging.getLogger(__name__)


def create_esp() -> Esp:
    try:
        return SerialEsp()
    except:
        log.warning('could not create serial esp')
        pass
    try:
        return WebEsp()
    except:
        log.warning('could not create web esp')
        pass
    return MockedEsp()



from __future__ import annotations
from asyncio.exceptions import CancelledError
from rosys import factory
import asyncio
import inspect
import logging
from typing import Awaitable, Callable, Optional, Type, Union, get_type_hints
from . import task_logger
from .persistence import Persistence
from .actors.actor import Actor
from .actors.odometer import Odometer
from .actors.steerer import Steerer
from .actors.automator import Automator
from .world.world import World, WorldState
from .world.mode import Mode
from . import event


class Runtime:

    def __init__(self, world: Optional[World] = None, persistence: Optional[Persistence] = None):
        self.world = world or World()
        Actor.world = self.world
        self.tasks = []
        self.log = logging.getLogger(__name__)

        if self.world.mode != Mode.TEST:
            self.persistence = persistence or Persistence(self.world)
            self.persistence.restore()

        self.esp = factory.create_esp()
        self.log.info(f'selected {type(self.esp).__name__}')
        self.odometer = Odometer()
        self.steerer = Steerer(self.esp)
        self.automator = Automator()

        self.actors = [
            self.esp,
            self.odometer,
            self.steerer,
            self.automator,
        ]

    def with_actors(self, *actors: list[Actor]):
        self.actors += actors
        return self

    async def pause(self, because: Optional[str] = None):
        if self.world.state == WorldState.PAUSED:
            return
        if because:
            await event.call(event.Id.NEW_NOTIFICATION, f'pausing automations because {because}')
        self.world.state = WorldState.PAUSED
        await self.esp.drive(0, 0)

    def resume(self):
        self.world.state = WorldState.RUNNING

    async def start(self):
        if self.tasks:
            raise Exception('should be only executed once')

        event.register(event.Id.NEW_NOTIFICATION, self.store_notification)
        event.register(event.Id.PAUSE_AUTOMATIONS, self.pause)
        for actor in self.actors:
            if actor.interval is not None:
                self.tasks.append(task_logger.create_task(self.repeat(actor), name=actor.name))
        self.tasks.append(asyncio.create_task(self.watch_emitted_events()))

        await asyncio.sleep(1)  # NOTE we wait for RoSys to start up before analyzing async debugging
        self.activate_async_debugging()

    async def stop(self):
        await self.esp.drive(0, 0)
        if self.world.mode != Mode.TEST:
            self.persistence.backup()
        [t.cancel() for t in self.tasks]
        Actor.process_pool.shutdown()
        await asyncio.gather(*[task_logger.create_task(a.tear_down()) for a in self.actors])

    async def repeat(self, actor: Actor):
        while True:
            start = self.world.time
            try:
                await actor.step()
                dt = self.world.time - start
            except (CancelledError, GeneratorExit):
                return
            except:
                dt = self.world.time - start
                self.log.exception(f'error in {actor}')
                if actor.interval == 0 and dt < 0.1:
                    delay = 0.1 - dt
                    self.log.warning(
                        f'{actor} would be called to frequently ' +
                        f'because it only took {dt*1000:.0f} ms; ' +
                        f'delaying this step for {delay*1000:.0f} ms')
                    await actor.sleep(delay)
            try:
                await actor.sleep(actor.interval - dt)
            except (CancelledError, GeneratorExit):
                return

    def get_actor(self, _type: Type):
        return next((a for a in self.actors if type(a) is _type), None)

    def store_notification(self, message: str):
        self.log.info(message)
        self.world.notifications.append((self.world.time, message))

    def activate_async_debugging(self):
        '''Produce warnings for coroutines which take too long on the main loop and hence clog the event loop'''
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            loop.slow_callback_duration = 0.05 if self.world.mode == Mode.REAL else 0.05
        except:
            self.log.exception('could not activate async debugging')

    async def watch_emitted_events(self):
        while True:
            try:
                for task in event.tasks:
                    if task.done() and task.exception():
                        self.handle_exception(task.exception())
                # cleanup finished tasks
                event.tasks = [t for t in event.tasks if not t.done()]
            except:
                self.log.exception('failing to watch emitted events')
            await asyncio.sleep(0 if self.world.mode == Mode.TEST else 0.1)

    def handle_exception(self, ex: Exception):
        self.log.exception(ex)
#        self.log.exception('task failed to execute', exec_info=ex)
